from junoplatform.io.utils import junoconfig
import yaml
import os
import logging
import requests
import json


def info_package(package_id:str=""):
    '''info packages
    returns: 
        tuple(code, data)
            code: 0 - success; otherwise - failure
            data: when(code==0) - dict; otherwise - errmsg(str)
    '''
    api = f'{junoconfig["cloud"]["api"]}/package/info'
    params = {}
    if not package_id:
        package_id = junoconfig["package_id"]
        
    params["package_id"] = package_id
    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {junoconfig['cloud']['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages"
        if "detail" in r.json():
            msg += ". detail: " + r.json()["detail"]
        return 1, msg
    else:
        data = r.json()
        data["config"] = json.loads(data["config"])
        #data["status"] = statusmap.get(data["status"])
        return 0, data

def list_packages(plant, module):
    '''list packages
    returns: 
        tuple(code, data)
            code: 0 - success; otherwise - failure
            data: when(code==0) - list[dict]; otherwise - errmsg(str)
    '''
    api = f"{junoconfig['cloud']['api']}/packages"
    params = {}
    if plant:
        params["plant"] = plant
    if module:
        params["module"] = module
    
    logging.info(f"list packages with params:\n{params}")
    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {junoconfig['cloud']['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return 1, msg
    else:
        res = []
        for x in r.json():
            x["config"] = json.loads(x["config"])
            #x["status"] = statusmap.get(x["status"])
            res.append(x)
        res.reverse()
        return 0, res

def deploy_package(package_id:str, kind:int):
    '''deploy a package
    kind: int, 0 - normal; 1 - rollback
    returns: 
        tuple(code, data)
            code: 0 - success; otherwise - failure
            data: when(code==0) - None; otherwise - errmsg(str)
    '''
    api = f"{junoconfig['cloud']['api']}/deploy"
    params = {}
    params["package_id"] = package_id
    params["kind"]= kind
    r = requests.post(api, params=params, headers={"Authorization": f"Bearer {junoconfig['cloud']['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return 1, msg
    else:
        return 0, None

def rollback(package_id:str=""):
    '''rollback a package to previous version or specific id[optional]
    returns: 
        tuple(code, data)
            code: 0 - success; otherwise - failure
            data: when(code==0) - new package_id(str); otherwise - errmsg(str)
    '''
    token = junoconfig['cloud']['token']
    if not package_id:
        try:
            package_id = junoconfig["package_id"]
        except:
            logging.error("no package id")
            exit(1)
    code, res = info_package(package_id=package_id)
    if code:
        logging.error(f"failed to get package info: {str}")
        return 1, res
    else:
        code, res = list_packages(res["plant"], res["module"])
        if code:
            logging.error(res)
            return 2, res
        else:
            res.reverse()
            target_idx = -1
            for idx, x in enumerate(res):
                if x["package_id"] == package_id:
                    target_idx = idx +1

            if target_idx < len(res):
                new_id = res[target_idx]["package_id"]
                code, res = deploy_package(new_id , 1)
                if not code:
                    logging.info(f"rollback from {package_id} to {new_id} submitted")
                    return 0, {new_id}
                else:
                    logging.error(res)
                    return 3, res
            else:
                msg = f"no available package to rollback for {package_id}"
                logging.error(msg)
                return 4, msg
        
    


from distutils.core import setup, Extension

def main():
    setup(name="fputs",
          version="1.0.0",
          description="Python interface for the fputs C library function",
          author="<your name>",
          author_email="your_email@gmail.com",
          ext_modules=[Extension("fputs", ["test_module.c"])])

if __name__ == "__main__":
    main()


"""junoplatform.io: implements io spec and common tools for algo"""
__author__      = "Bruce.Lu"
__email__       = "lzbgt@icloud.com"
__time__ = "2023/07/20"

__all__ = [
    'Storage',
    'InputConfig',
]

from pydantic import BaseModel, Field, model_validator, NonNegativeInt
import datetime as datetime
from typing import Optional, Any, List
import numpy as np

from junoplatform.io._driver import Pulsar as Pulsar, Opc as Opc, junoconfig, Redis, Clickhouse, RDLock
from junoplatform.io.utils import junoconfig
import logging


class InputConfig(BaseModel):
    ''' InputConfig spec for JunoPlatform Runtime
    tags: OPC tags
    minutes: last n minutes of data
    items: last n records of data
    sched_interval: algo schedule interval in seconds
    '''
    tags: List[str]
    minutes: Optional[NonNegativeInt] = Field(default=None, description='input data of last n minutes')
    items: Optional[NonNegativeInt] = Field(default= None, description='input data of last n items')
    sched_interval: NonNegativeInt = Field(description='schedule interval in seconds')

    @model_validator(mode="before")
    def atleast_one(cls, values: 'dict[str, Any]') -> 'dict[str, Any]':
        if not values.get('minutes') and not values.get('items'):
            raise ValueError("field 'minutes' or 'items' must be given")
        return values

class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]
    
class DataSet():
    def __init__(self) -> None:

        if 'clickhouse' in junoconfig:
            self.url = junoconfig['clickhouse']['url']
        else:
            devcfg = {'gangqu': 'ch://sf_read1:reader@123.@192.168.101.101:7000', 
                      'yudai': 'ch://default:L0veclickhouse@192.168.101.101:8123',
                      'yulin':'ch://sf_read1:reader@123.@192.168.101.101:7010'}
            self.url = devcfg[junoconfig["plant"]]
        self.io:Clickhouse = Clickhouse(url=self.url)
        self.tbl_c = f'{junoconfig["plant"]}_data.running'
        self.tbl_w = f'{junoconfig["plant"]}_data.running_today'

    def fetch(self, num:int = 0, tags:List[str] = "", time_from = None):
        if num >0:
            data, timestamps, names = self.io.read(self.tbl_w, tskey='time', tags=tags, num = num, time_from=None)
            num = num - len(timestamps)
            if num > 0:
                nd, nt, _ = self.io.read(self.tbl_c, tskey='time', tags=tags, num = num, time_from=None)
                return np.concatenate((nd, data), axis=0), nt + timestamps, names
            return data, timestamps, names
        if time_from:
            data, timestamps, names = self.io.read(self.tbl_w, tskey='time', tags=tags, num = 0, time_from=time_from)
            if timestamps[0] > time_from:
                nd, nt, _ = self.io.read(self.tbl_c, tskey='time', tags=tags, num = 0, time_from=time_from)
                return np.concatenate((nd, data), axis=0), nt + timestamps, names
            return data, timestamps, names


class Storage(metaclass=Singleton):
    _cloud = None
    _opc = None
    _local = None

    def __init__(self):
        super(Storage, self).__init__()
        _redis_cfg = junoconfig['redis']
        self._local = Redis(**_redis_cfg)
        self.lock = RDLock(self._local.io)

    @property
    def cloud(self):
        if not self._cloud:
            CloudStorage:Pulsar = Pulsar(**junoconfig['pulsar'], lock=self.lock)
            self._cloud = CloudStorage
        return self._cloud
    
    @property
    def opc(self):
        if not self._opc:     
            OpcWriter = Opc(lock=self.lock)
            self._opc = OpcWriter  
        
        return self._opc
    
    @property
    def local(self):
        return self._local



import requests
import logging

def write(data:list[dict]):
    """
    data: list[dict], example:
          [{'PLC_TAG1': 2.2}, {'PLC_TAG2': 1.2}]
    """
    try:
      api='http://jp-connector/api/write'
      r = requests.post(api, json=data, timeout=5)
      if r.status_code != 200:
         logging.error(r.text)
         return r.text
      logging.info(f"write opc success: {data}")
      return ""
    except Exception as e:
      logging.error(str(e))
      return str(e)


"""junoplatform.io._driver.py: implements DB adaptor classes"""
__author__      = "Bruce.Lu"
__email__       = "lzbgt@icloud.com"
__time__ = "2023/07/20"

from abc import ABC, abstractmethod
import itertools
from junoplatform.io.utils import *
import logging
from typing import Dict
from pulsar import Client, AuthenticationTLS, ConsumerType, InitialPosition, schema, Producer, Consumer
import json
from datetime import datetime
from typing import List
import dateparser
import numpy as np
from junoplatform.api.local import write as connector_write
import redis

class IWriter(ABC):
    @abstractmethod
    def write(self, **kwargs):
        pass

class IReader(ABC):
    @abstractmethod
    def read(self, **kwargs):
        pass

class ALock(ABC):
    @abstractmethod
    def aquire(self, key:str, ex:int):
        pass


class RDLock(ALock):
    def __init__(self, redis:redis.Redis):
        super(RDLock, self).__init__
        self.redis = redis
        self.key = f"{junoconfig['plant']}.{junoconfig['module']}.rdlock"
        marg = junoconfig['input_cfg']["sched_interval"]/3
        self.ex = junoconfig['input_cfg']["sched_interval"] + marg
        self.uid:str = junoconfig['instance_id']

    def aquire(self):
        try:
            if "." not in self.uid:
                return True
            
            r = self.redis.setnx(self.key, self.uid)
            if not r:
                v = self.redis.get(self.key)
                if v.decode() != self.uid:
                    return False
            self.redis.expire(self.key, self.ex)
            return True
        except Exception as e:
            logging.error({e})
            return False

class Opc(IWriter, IReader):
    def __init__(self, lock:ALock=None ,**kwargs,):
        super(Opc, self).__init__()
        self.io = kwargs
        self.lock = lock
    
    def write(self, data:list[dict], **kwargs):
        if self.lock:
            if not self.lock.aquire():
                logging.warn("holding no lock, skip opc write")
                return
        return connector_write(data)

    def read(self, **kwargs):
        pass

class Pulsar(IWriter, IReader):
    def __init__(self, lock:ALock = None,  **kwargs):
        super(Pulsar, self).__init__()
        self.io = pulsar_cli(**kwargs)
        logging.info(f"pulsar args: {kwargs}")
        self.producers: Dict[str, Producer] = {}
        self.consumers: Dict[str, Consumer] = {}
        self.plant = junoconfig["plant"]
        self.module = junoconfig["module"]
        self.instance_id = junoconfig["instance_id"]
        # prevent multipule instance issue
        self.name = f"algo-{self.plant}-{self.module}-{self.instance_id}"
        self.lock=lock

    def write(self, table: str, data:dict, **kwargs):
        if self.lock:
            if not self.lock.aquire():
                logging.warn("holding no lock, skip pulsar write")
                return
            
        topic = f'up-{self.plant}-{self.module}_{table}'
        if not (topic in self.producers):
            self.producers[topic] = self.io.create_producer(topic)

        self.producers[topic].send(json.dumps(data).encode('utf-8'))
        logging.info(f'write cloud: sent {data} to {topic}')
        
    def read(self, topic:str, shared:bool=False):
    #   if self.lock:
    #         if not self.lock.aquire():
    #             logging.warn("holding no lock, skip pulsar read")
    #             return
            
      subtype = ConsumerType.Shared
      if not shared:
          subtype = ConsumerType.Exclusive

      if topic not in self.consumers:
        
        self.consumers[topic] = self.io.subscribe(topic, self.name, 
                              consumer_type=subtype, initial_position=InitialPosition.Latest, 
                              schema=schema.BytesSchema(), pattern_auto_discovery_period=1, broker_consumer_stats_cache_time_ms=1500)
        
      return self.consumers[topic].receive()

class Mongo(IWriter, IReader):
    def __init__(self, **kwargs):
        super(Mongo, self).__init__()
        self.url = kwargs['url']
        self.io = mongo_cli(self.url)
    
    def write(self, table:str, data:dict, **kwargs):
        ''' write to mongodb
            table: str, collection name
            data: dict, document to store
        database is derived from runtime environment
        '''
        logging.info(junoconfig)

        plant = junoconfig['plant']
        module = junoconfig['module']
        self.io[plant][f'{module}_{table}'].insert_one(data)

    def read(self, **kwargs):
        pass

class Redis(IWriter, IReader):
    def __init__(self, **kwargs):
        super(Redis, self).__init__()
        self.io = redis_cli(**kwargs)
        self.lock = RDLock(self.io)

    def write(self, key: str, value:dict|bytes, isBytes = False, **kwargs):
        if not self.lock.aquire():
            logging.warn("holding no lock, skip redis write")
            return
        
        if isBytes and not isinstance(value, bytes):
            raise Exception("invalid value type")
        if not isBytes:
            logging.info(f"write local:  {key} -> {value}")
            self.io.json().set(key, '$', value)
        else:
            self.io.set(key, value)

    def read(self, key:str, isBytes = False, **kwargs):
        if not isBytes:
            return self.io.json().get(key)
        else:
            return self.io.get(key)

# class Elastic(IWriter, IReader):
#     def __init__(self, **kwargs):
#         super(Elastic, self).__init__()
#         self.url = kwargs['url']
#         self.ca = kwargs['ca']
#         self.user=kwargs['user']
#         self.password=kwargs['password']
#         self.io:Elasticsearch=es_cli(self.url, ca=self.ca,user=self.user, password=self.password)
    
#     def write(self, **kwargs):
#         topic=kwargs['topic']
#         data=kwargs['data']
#         id = ""

#         if "et" in data:
#             id = str(data["et"])
        
#         for x in ["time", "ts", "timestamp"]:
#             if id:
#                 break
#             if x in data:
#                 if isinstance(data[x], str):
#                     try:
#                         id = str(dateparser.parse(data[x]).timestamp()*1000)
#                     except:
#                         pass
#                 elif isinstance(data[x], datetime):
#                         id = str(data[x].timestamp()*1000)
            
#         if id:
#             self.io.index(index=topic, id=id, document=data)
#         else:
#             self.io.index(index=topic, document=data)
    
#     def read(self, **kwargs):
#         pass

class Clickhouse(IWriter, IReader):
    def __init__(self, *args, **kwargs):
        super(Clickhouse, self).__init__()
        self.url = kwargs['url']
        self.io: CHClient = clickhouse_cli(self.url)
    
    def write(self, **kwargs):
        pass
    
    def read(self, table:str,  tskey: str, tags:List[str]=[], num=0, time_from = None, time_to = None):
        cols = "*"
        if tags:
            if tskey not in tags:
                tags.append(tskey)
            tags = list(map(lambda x: f"`{x}`", tags))    
            cols = ", ".join(tags)

        sql = f"select {cols} from {table}"
        for x in [time_from, time_to]:
            if isinstance(x, str) or isinstance(x, datetime) or x is None:
                pass
            else:
                raise Exception("invalid time_from or time_to")

        if not time_to and not time_from and not num:
            raise Exception("must provide time_from or time_to or num")
        else:
            s1 = ""
            s2 = ""
            if time_to:
                if isinstance(time_to, datetime):
                    time_to = time_to.strftime("%Y-%m-%d %H:%M:%S")
                s1 = f"{tskey} <= '{time_to}'"
            if time_from:
                if isinstance(time_from, datetime):
                    time_from = time_from.strftime("%Y-%m-%d %H:%M:%S")
                s2 = f"{tskey} > '{time_from}'"
            if s1 and s2:
                sql += f" where {s1} and {s2} "
            else:
                for x in [s1, s2]:
                    if x:
                        sql += f" where {x} "
        order = "asc"
        if num > 0:
            order = "desc"

        sql += f" order by {tskey} {order}"
        if num > 0:
            sql+=f" limit {num}"

        logging.info(sql)
        try:
            r = self.io.query(sql)
            logging.info(f"got {len(r.result_rows)} records from dataset")
            data = list(map(lambda x: x[:-1], r.result_rows))
            timestamps = [x[-1] for x in r.result_rows]
            data = np.array(data,dtype=np.float32)
            column_names = r.column_names[:-1]

            if num > 0:
                timestamps.reverse()
                data = np.flip(data, 0)

            return data, timestamps, column_names
        except Exception as e:
            logging.error(f"exception: {e}")
            return None, None, None

# class Qdb(IWriter, IReader):
#     def __init__(self, *args, **kwargs):
#         super(Qdb, self).__init__()
#         self.init_kwargs=kwargs
#         self.io = qdb_cli(**self.init_kwargs)
    
#     def write(self, **kwargs):
#         topic=kwargs['topic']
#         data={k.translate(k.maketrans({'-':'_', '.': '_'})):v  for k,v in kwargs['data'].items()}
#         buff = self.io.new_buffer()
#         buff.row(topic, columns=data)
#         self.io.flush(buff)
    
#     def read(self, **kwargs):
#         pass

def MakeDB(kind: str, **kwargs):
    if kind == 'clickhouse':
        return Clickhouse(**kwargs)
    # elif kind == 'elastic':
    #     return Elastic(**kwargs)
    # elif kind == 'qdb':
    #     return Qdb(**kwargs)
    elif kind == 'mongo':
        return Mongo(**kwargs)
    else:
        raise Exception(f'unkown kind of db: {kind}')


import logging
from junoplatform.io import Storage
from junoplatform.io.utils import junoconfig

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(filename)s %(lineno)d - %(message)s')

class RTLogHandler(logging.Handler):
    def __init__(self):
        super(RTLogHandler, self).__init__()
        self.io = Storage().cloud

    def emit(self, record):
        log_entry = {
            'level': record.levelname,
            'message': record.getMessage(),
            'et': record.created,
            'logger_name': record.name,
            'file': record.filename,
            'line_number': record.lineno,
            'function_name': record.funcName
        }

        self.io.write('rtlog', log_entry)

try:
    logger = logging.getLogger('rtlog')
    logger.addHandler(RTLogHandler())
    logger.setLevel(logging.INFO)
except:
    pass

    
    
    


"""decorators.py: decorator class and functions for junoplatform"""

import logging
from collections import deque
from urllib.parse import parse_qs
import traceback
import json
import os
import time
import datetime
from functools import wraps
from junoplatform.log import logger
from junoplatform.io.utils import JunoConfig, junoconfig
from junoplatform.io import InputConfig, Storage, DataSet
from threading import Thread
import numpy as np
__author__ = "Bruce.Lu"
__email__ = "lzbgt@icloud.com"
__time__ = "2023/07/20"


class EntryPoint:
    def __init__(self, cfg_in: str | InputConfig, detached: bool = False):
        super(EntryPoint, self).__init__()
        self.cfg_in: InputConfig
        self.detached = detached
        self.storage = Storage()
        self.dataset = DataSet()
        self.que = deque(maxlen=2000)
        self.stop_flag = False
        if isinstance(cfg_in, str):
            logger.debug(f"loading input spec from file: {cfg_in}")
            try:
                self.cfg_in = InputConfig(**json.load(open(cfg_in)))
            except Exception as e:
                msg = f"error in input.json: {e}"
                logger.error(msg)
                exit(1)
        elif isinstance(self.cfg_in, InputConfig):
            logger.info(f"loading input spec from class: {cfg_in}")
            self.cfg_in = cfg_in
        else:
            raise Exception(
                f"cfg_in must be type of InputConfig or string, but provides: {type(self.cfg_in)}")

    def _thread(self, func):
        while True:
            delay = self.cfg_in.sched_interval
            if not self.stop_flag:
                newcfg = JunoConfig()
                algo_cfg = newcfg["algo_cfg"]
                logger.info(f"running algo with junoconfig {newcfg}")

                ts = datetime.datetime.now().timestamp()
                if self.cfg_in.items:
                    data, timestamps, names = self.dataset.fetch(
                        tags=self.cfg_in.tags, num=self.cfg_in.items)
                elif self.cfg_in.minutes:
                    time_from = datetime.datetime.now() - datetime.timedelta(minutes=self.cfg_in.minutes)
                    data, timestamps, names = self.dataset.fetch(
                        tags=self.cfg_in.tags, time_from=time_from)
                else:
                    raise Exception("invalid InputConfig")

                td = datetime.datetime.now().timestamp()
                logger.info(f"time used fetching dataset: {td-ts}s")

                try:
                    func(self.storage, algo_cfg, data, timestamps, names)
                except Exception as e:
                    msg = traceback.format_exc()
                    logger.error(f"{e}: {msg}")

                # TODO: output
                te = datetime.datetime.now().timestamp()
                logger.info(f"time used running algo: {te-td}s")

                delay = self.cfg_in.sched_interval - (te - ts) - 0.003
                logger.debug(
                    f"delay in seconds to make up a full sched_interval: {delay}")
                if delay < 0:
                    delay = 0
            else:
                logging.info("module disabled, skip run algo func")

            time.sleep(delay)

    def s_get_bool(self, v: str):
        try:
            x = int(v)
            return x != 0
        except:
            if v.lower() in ["t", "true", "y", "yes", "ok", "on", "enable", "active"]:
                return True
            else:
                return False

    def _pulsar(self):
        itopic = f"jprt-down-{junoconfig['plant']}-{junoconfig['module']}"
        while True:
            msg = self.storage.cloud.read(itopic, shared=False)
            data = {}
            try:
                data = json.loads(msg.data())
                if "package_id" not in data or junoconfig['package_id'] != data["package_id"]:
                    logger.error(
                        f"invalid msg received {data}, self package_id: {junoconfig['package_id']}")
                    self.storage.cloud.consumers[itopic].acknowledge(msg)
                    continue
            except:
                logger.error(f"invalid msg received: {msg.data()}")
                self.storage.cloud.consumers[itopic].acknowledge(msg)
                continue

            if "cmd" in data:
                cmd = parse_qs(data['cmd'])
                v = cmd.get('enable', [''])[0]
                if v:
                    enable = self.s_get_bool(v)
                    if not enable:
                        self.stop_flag = True
                        logging.info("enable=false cmd received")
                    else:
                        self.stop_flag = False
                        logging.info("enable=true cmd received")

                    data = {"enable": enable,
                            "et": datetime.datetime.now().timestamp()*1000, "kind": "1", "package_id": data["package_id"]}
                    self.storage.cloud.write("module_state_new", data)
            else:
                logging.error(f"unkown msg: {data}")

            self.storage.cloud.consumers[itopic].acknowledge(msg)

    def _heart_beat(self):
        while True:
            data = {"enable": not self.stop_flag,
                    "et": datetime.datetime.now().timestamp()*1000, "kind": "0", "package_id": junoconfig["package_id"]}
            self.storage.cloud.write("module_state_new", data)

            # heartbeat interval is 1 minutes
            time.sleep(60)

    def __call__(self, func):
        pt = Thread(target=self._pulsar)
        pt.start()

        hb = Thread(target=self._heart_beat)
        hb.start()

        th = Thread(target=self._thread, args=(func,))
        if self.detached:
            th.daemon = True
        th.start()
        if not self.detached:
            th.join()


def auth(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if not os.path.exists(args[0].juno_dir) or not os.path.exists(args[0].juno_file):
            logger.error(f"user not authenticationd.\n\
                          please run `junocli login [api_url]` to use your shuhan account")
            os.makedirs(args[0].juno_dir, exist_ok=True)
            return -1
        return func(*args, **kwargs)

    return wrapper



"""junoplatform.io.utils.py: implements helper tools"""
__author__      = "Bruce.Lu"
__email__       = "lzbgt@icloud.com"
__time__ = "2023/07/20"


from elasticsearch import Elasticsearch
import clickhouse_connect
import pymongo
from questdb.ingress import Sender, IngressError, TimestampNanos, TimestampMicros
import oss2
from itertools import islice
import pulsar
import re
import sys
import traceback
import time
import uuid
import logging
from urllib.parse import urlparse
from clickhouse_connect.driver import Client as CHClient
import yaml
import redis
from datetime import timedelta
import uuid
import json
from decouple import config as dcfg
import os
from collections import UserDict

def get_package_path(cfg: dict, package_id: str):
    plant = cfg['plant']
    module = cfg['module']
    
    return f"dist/{plant}-{module}-{package_id}.zip"

class JunoConfig(UserDict):
    def __init__(self):
        super(JunoConfig, self).__init__()
        self.data = {}
        if os.path.exists('project.yml'):
            with open('project.yml', 'r') as f:
                self.data.update(yaml.safe_load(f))
        
        if os.path.exists('project-ext.yml'):
            with open('project-ext.yml', 'r')  as f:
                ext = yaml.safe_load(f)
                if ext and len(ext) > 0:
                    self.data.update()

        if "cloud" in self.data and "token" in self.data["cloud"]:
            pass
        else:
            jf = os.path.expanduser('~') + '/.juno/config.yaml'
            if os.path.exists(jf):
                with open(jf, 'r') as f:
                    jc = yaml.safe_load(f)
                    self.data["cloud"] = jc
            else:
                self.data["cloud"] = {"api": "http://192.168.101.158:/8823/api"}
                    
        if "redis" not in self.data:
            self.data["redis"] = {
                "host": "192.168.101.158",
                "port": 6379,
                "password": "myredis"
            }
        if "pulsar" not in self.data:
            self.data["pulsar"] = {
                "url": "pulsar://192.168.101.158:6650"
            }

        
        self.data["instance_id"] = instance_id = dcfg("instance_id", "") if dcfg("instance_id", "") else uuid.uuid4().hex
        self.data["input_cfg"] = {}
        try:
            with open('input.json', 'r') as f:
                self.data["input_cfg"] = json.load(f)
        except:
            pass

        self.data["algo_cfg"] = {}
        try:
            with open('config.json', 'r') as f:
                self.data["algo_cfg"] = json.load(f)
        except:
            pass

        if dcfg("cloud_api", ""):
            if "cloud" in self.data:
                self.data["cloud"]["api"] = dcfg("cloud_api")
            else:
                self.data["cloud"] = {"api":dcfg("cloud_api")}

    def __getitem__(self, key):
        try:
            return self.data[key]
        except KeyError:
            pass
        try:
            return self.data[key]
        except KeyError:
            pass
        raise KeyError(key)

    def __setitem__(self, key, value):
        self.data[key] = value

junoconfig = JunoConfig()

def redis_cli(host: str, port: int, password:str, db:int=0, socket_timeout = 3):
    logging.debug(f"local redis: {host}:{port} {db} {password}")
    return redis.Redis(host, port, db, password, socket_timeout=socket_timeout)
     
def pulsar_cli(url:str, shared: bool = False, ca:str="certs/ca.cert.pem",
               cert:str="certs/client.cert.pem", key:str="certs/client.key-pk8.pem"):
    client: pulsar.Client
    if 'ssl' in url:
        auth = pulsar.AuthenticationTLS(cert, key)
        client = pulsar.Client(url,
                        tls_trust_certs_file_path=ca,
                        tls_allow_insecure_connection=False,
                        authentication=auth)
    else:
        client = pulsar.Client(url)
        
    return client


def es_cli(url:str, ca:str, user:str, password:str):
    #cfg['elastic']['url'], ca_certs=cfg['elastic']['ca'], 
    #basic_auth=(cfg['elastic']['user'], cfg['elastic']['password'])
    return Elasticsearch(url, ca_certs=ca, basic_auth=(user, password))


def clickhouse_cli(url: str):
        p = urlparse(url)
        schema = p.scheme
        if schema != 'ch':
            raise Exception(f'invalid schema in dbs.clickhouse.url: {url}')
        user = p.username
        password = p.password
        host = p.hostname
        port = p.port
        return clickhouse_connect.get_client(host=host, username=user, 
                                            password=password, port=port)

# auth = ("pulsar", "KA6oqjb0s5OP49WHfKLabO8ef42ArV_9q9NznHNKUJ8", "_lTCuKKqtRWVDyb9d5545s99VwXVkJjs-HhCtPxPaTQ", "afh08JDKKYUxPYuHCWsytHQ7LgUZ63s-CTvTFaWFVE4")
def qdb_cli(host:str, port: str, auth: str, tls: bool, auto_flush: bool, **kwargs):
     auth_t = tuple(auth.split(' '))
     s = Sender(host, port, auth=auth_t, tls=tls, auto_flush=auto_flush)
     s.connect()
     return s
          

def mongo_cli(url: str):
    return pymongo.MongoClient(url)

def oss_cli(key:str, sec:str, endpoint:str, bucket: str):
    auth = oss2.Auth(key, sec)
    endpoint = endpoint
    r = oss2.Bucket(auth, endpoint, bucket, connect_timeout=5)
    # for b in islice(oss2.ObjectIterator(r), 10):
    #     print(b.key)
    return r



from junoplatform.meta.decorators import EntryPoint
from junoplatform.io import *
from junoplatform.log import logger
import random
import json

#@EntryPoint(InputConfig(tags=["AI-T20502_DIS", "AI-T20501_DIS"], items=1440, interval=5), detached=False)
@EntryPoint("input.json", detached=False)
def any_algo_entry_func(storage:Storage, algo_cfg, data, timestamps, names):
    ''' params signature of algo func:
    storage: Storage
      framework provided storage object
    algo_cfg: dict
      algo configration from config.json
    data: numpy.ndarray
      framework provided input dataset which specified by `InputConfig` or input.json
    timestamps: List[datetime.datetime]
      timestamps for data
    names: List[str]
      col names for data
    '''

    # demo: algo processing with input data here
    logger.info(f"processing data: {data.shape}, tags: {names}, time: {timestamps[0]} ~ {timestamps[-1]}")
    
    # demo: construct results as list of dict: [{"key": key, "value": value}]
    opc_data = [{"key": f"{names[0]}", "value": random.randint(1,10)}, {"key": f"{names[1]}", "value": random.randint(1,10)}]
    probe1 = {"pX":random.randint(1,10)}
    state1 = {'pY': 1}

    # demo: algo oputput results
    logger.info("data processed, writing outputs:")
    # write data to cloud
    storage.cloud.write("probe1", probe1)

    # write data to opc
    storage.opc.write(opc_data)

    # read/write data to local
    storage.local.write('state1_bytes', json.dumps(state1).encode(), isBytes=True)
    data = storage.local.read('state1_bytes', isBytes=True)
    logger.info(f"read write bytes: {data}")

    storage.local.write('state1_dict', state1)
    data = storage.local.read('state1_dcit')
    logger.info(f"read write dict: {data}")


"""junoplatform.tools.cmd.py: junocli cmd implementation"""
__author__      = "Bruce.Lu"
__email__       = "lzbgt@icloud.com"
__time__ = "2023/07/20"


import click
import os
import yaml
import logging
import requests
import shutil
import traceback
import yaml
import uuid
import json
from io import StringIO
from pylint.lint import Run
from pylint.reporters.text import TextReporter
from typing import List
import zipfile
from junoplatform.io.utils import junoconfig, get_package_path

import junoplatform
from junoplatform.meta.decorators import auth
from typing import Optional, Mapping
import collections
from decouple import config as dcfg

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s - %(message)s')
        
class CMDBase(object):
    def __init__(self):
        self.juno_dir = os.path.expanduser('~') + '/.juno'
        os.makedirs(self.juno_dir, exist_ok=True)
        self.juno_file = self.juno_dir +  '/config.yaml'
        self.juno_cfg = {}
        try:
            self.juno_cfg = yaml.safe_load(open(self.juno_file, 'r'))
        except:
            pass

class OrderedGroup(click.Group):
    def __init__(self, name: Optional[str] = None, commands: Optional[Mapping[str, click.Command]] = None, **kwargs):
        super(OrderedGroup, self).__init__(name, commands, **kwargs)
        #: the registered subcommands by their exported names.
        self.commands = commands or collections.OrderedDict()

    def list_commands(self, ctx: click.Context) -> Mapping[str, click.Command]:
        return self.commands

@click.group(cls=OrderedGroup)
@click.pass_context
def main(ctx, ):
    ctx.obj = CMDBase()

pass_base = click.make_pass_decorator(CMDBase)


@main.command()
@click.argument("api")
@click.option("-u", "--username", prompt=True)
@click.option("-p", "--password", prompt=True, hide_input=True,
)
@pass_base
def login(base:CMDBase, api, username, password):
    '''must login success before all other commands
    '''
    auth = {"username": username, "password": password}
    logging.info(f"login at {api}")
    r = requests.post(f'{api}/login', data=auth, headers = {'Content-Type': 'application/x-www-form-urlencoded'})
    if r.status_code != 200:
        if 'detail' in r.json():
            detail = r.json()['detail']
            logging.error(f"login error: {detail}")
            return
        else:
            logging.error(f"login error: {r.status_code}")
    token = r.json()['access_token']
    data = {"auth": auth, "token": token, "api": api}

    with open(base.juno_file, 'w') as f:
        f.write(yaml.dump(data)) 
    logging.info("successfully logged in")

@main.command()
@click.argument('name')
@click.argument('plant')
@click.argument('module')
@pass_base
@auth
def create(base:CMDBase, name, plant, module):
    '''create an algo module with project NAME for PLANT-MODULE
    '''
    home = os.path.dirname(junoplatform.__file__)
    src = f"{home}/templates/main.py"
    try:
        os.makedirs(name, exist_ok=False)
        shutil.copy2(src, name)
        doc = {"name": name, "version": "0.0.0", "author": os.getlogin(), "package_id": "int_temporary",
                "description": "template algo project", "plant": plant, "module": module}
        with open(f"{name}/project.yml", "w") as f:
            yaml.dump(doc, f, sort_keys=False)
        input = {
        "tags": [
            "AI-T20502_DIS",
            "AI-T20501_DIS"
        ],
        "items": 1440,
        "sched_interval": 90
        }
        
        if plant == 'yudai':
            input["tags"] = ['AI_AT-P10503-DISP', 'COD2']
        
        if plant == 'yulin':
            input['tags'] = ['通道 1.设备 1.opc.group.3004FV00306_AI.PV', '通道 1.设备 1.opc.group.3007LIT00202_AI.PV']

        config = {"key1": 1, "key2": [1,2]}
        with open(f"{name}/input.json", "w") as f:
            json.dump(input, f, sort_keys=False)

        with open(f"{name}/config.json", "w") as f:
            json.dump(config, f, sort_keys=False)
    except Exception as e:
        msg = traceback.format_exc()
        logging.error(f"failed to create project {name}: {e}")

@main.command()
@pass_base
@auth
def run(base):
    '''run a package locally for testing
    '''
    os.system("python main.py")

@main.group(cls=OrderedGroup)
@click.pass_context
def package(ctx):
    pass

@main.group(cls=OrderedGroup)
@click.pass_context
def deploy(ctx):
    pass

@main.group(cls=OrderedGroup)
@click.pass_context
def plant(ctx):
    pass

@package.command()
@click.argument('conf_file', default="config.json")
@click.option('-t', '--tag', type=click.Choice(['algo', 'config', 'ac', 'other']), required = True)
@click.option('-m', '--message', required = True)
@click.option('-i','--input', help = "the path of input spec file", default="input.json")
@pass_base
@auth
def new(base:CMDBase, conf_file, tag, message, input):
    ''' make a new package
    '''
    try:
        lint = StringIO()  # Custom open stream
        reporter = TextReporter(lint)
        Run(["main.py"], reporter=reporter, exit=False)
        errors = lint.getvalue().split("\n")
        for x in errors:
            if "failed"  in x or "fatal" in x:
                logging.error(x)
                logging.info("fix the error above and redo package")
                exit(1)

        package_id = uuid.uuid4().hex
        junoconfig['package_id'] = package_id

        def parse_version(s:str) -> List[int]|None:
            v = s.split(".")
            if len(v) != 3:
                return None
            try:
                return [int(x) for x in v]
            except:
                return None
            
        def inc_version(v: str, t:type):
            v = parse_version(v)
            if tag == 'algo':
                v[0] +=1
                junoconfig['tag'] = 'algo'
            elif tag == 'config':
                v[1] +=1
                junoconfig['tag'] = 'config'
            elif tag == 'ac':
                v[0] +=1
                v[1] +=1
                junoconfig['tag'] = 'ac'
            else:
                v[2] +=1
            junoconfig["version"] = ".".join([str(x) for x in v])

        if "version" not in junoconfig:
            junoconfig["version"] = "0.0.0"
        try:
            inc_version(junoconfig["version"], tag)
        except:
            logging.error(f"invalid version: {junoconfig['version']}")
            exit(1)

        junoconfig["message"] = message


        logging.info(f"junocofig: {junoconfig.__dict__}")

        if "algo_cfg" in junoconfig:
            junoconfig.pop("algo_cfg")
        if "input_cfg" in junoconfig:
            junoconfig.pop('input_cfg')
        if "cloud" in junoconfig and "auth" in junoconfig["cloud"]:
            junoconfig["cloud"].pop("auth")
            
        with open('project.yml', 'w') as f:
            yaml.safe_dump(junoconfig.data, f, sort_keys=False)

        logging.info(f"create new package success (updated project.yml):\n\tplant: {junoconfig['plant']}, \
                     module: {junoconfig['module']}, conf: {conf_file}\n\t{tag}: {message}\
                     \n\tid: {package_id}\n\tversion: {junoconfig['version']}")

        # dist
        os.makedirs("dist", exist_ok=True)
        module = junoconfig['module']
        plant = junoconfig['plant']
        arch = f'dist/{plant}-{module}-{junoconfig["package_id"]}.zip'
        with zipfile.ZipFile(arch, 'w') as f:
            for root, dirs, files in os.walk('./'):
                if root[-4:] == 'dist':
                    continue
                for file in files:
                    p = os.path.join(root, file)
                    f.write(p)
                    logging.info(f"added: {p}")
        logging.info(f"package stored in: {arch}")

    except Exception as e:
        msg = traceback.format_exc()
        logging.error(f"{e}:\n{msg}")

@package.command()
@click.argument('package_id', default="", required=False)
@pass_base
@auth
def pub(base:CMDBase, package_id):
    '''publish a package only (no deploy action)
    '''
    if not package_id:
        try:
            package_id = junoconfig["package_id"]
        except:
            logging.error("must under project dir to run this command when no package_id provided(using project default)")
            exit(1)
    
    api = f'{base.juno_cfg["api"]}/package'
    logging.info(f"build package {package_id} to {api}")
    papath = get_package_path(junoconfig, package_id)

    logging.info(base.juno_cfg['token'])
    with open(papath, 'rb') as f:
        r = requests.post(api, files = {'file': (f'{package_id}.zip', f, 'application/zip')}, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
        if r.status_code != 200:
            msg = f"faild upload package {package_id} "
            if "detail" in r.json():
                msg += r.json()["detail"]
            logging.error(msg)
        else:
            logging.info(f"successfully upload package {package_id}")

@package.command()
@click.argument('package_id', required=False)
@pass_base
@auth
def repub(base:CMDBase, package_id):
    if not package_id:
        try:
            package_id = junoconfig["package_id"]
        except:
            logging.error("must under project dir to run this command when no package_id provided(using project default)")
            exit(1)
    api = f'{base.juno_cfg["api"]}/package'
    params = {"package_id": package_id}
    r = requests.patch(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild rebuild package in cloud:  {package_id} "
        if "detail" in r.json():
            msg += r.json()["detail"]
        logging.error(msg)
    else:
        logging.info(f"successfully summit recompile package request: {package_id}")

statusmap = {
    0: "building",
    1: "ready",
    2: "failed"
}


def info_package(base:CMDBase, package_id:str):
    '''info packages
    '''
    api = f'{base.juno_cfg["api"]}/package/info'
    params = {}
    params["package_id"] = package_id
    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages"
        if "detail" in r.json():
            msg += ". detail: " + r.json()["detail"]
        return msg
    else:
        data = r.json()
        #data["config"] = json.loads(data["config"])
        #data["status"] = statusmap.get(data["status"])
        return data

@package.command()
@click.argument('package_id', required=False)
@pass_base
@auth
def info(base:CMDBase, package_id):
    '''info packages
    '''
    if not package_id:
        try:
            package_id = junoconfig["package_id"]
        except:
            logging.error("must under project dir to run this command when no package_id provided(using project default)")
            exit(1)

    res = info_package(base, package_id)
    if isinstance(res, str):
        logging.error(res)
    else:
        logging.info(json.dumps(res, indent=2))


@package.command()
@click.argument('package_id', required=True)
@pass_base
@auth
def clone(base:CMDBase, package_id):
    '''clone package
    '''
    if not dcfg("ENABLE_DL", 0, cast=int):
        logging.error("package download is not enabled")

    api = f"{base.juno_cfg['api']}/package"
    params = {}
    params["package_id"] = package_id
    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        logging.error(msg)
    else:
        with open(f"{package_id}.zip", "wb") as f:
            for c in r.iter_content(chunk_size=2048):
                if c:
                    f.write(c)

def list_packages(base:CMDBase, plant, module):
    '''list packages
    '''
    api = f"{base.juno_cfg['api']}/packages"
    params = {}
    if plant:
        params["plant"] = plant
    if module:
        params["module"] = module
    
    logging.info(f"list packages with params:\n{params}")
    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return msg
    else:
        res = []
        for x in r.json():
            #x["config"] = json.loads(x["config"])
            #x["status"] = statusmap.get(x["status"])
            res.append(x)
        res.reverse()
        return res

@package.command()
@click.argument('plant', required=False)
@click.argument('module', required=False)
@pass_base
@auth
def ls(base, plant, module):
    '''list packages
    '''
    res = list_packages(base, plant, module)
    if isinstance(res, str):
        logging.error(res)
    else:
        logging.info(f"results:\n{json.dumps(res, indent=2)}")

def deploy_package(base:CMDBase, package_id:str, kind:int):
    api = f"{base.juno_cfg['api']}/deploy"
    params = {}
    params["package_id"] = package_id
    params["kind"]= kind
    r = requests.post(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild deploy packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return msg
    else:
        return None

@deploy.command()
@click.argument('package_id', required=False)
@pass_base
@auth
def new(base, package_id):
    '''deploy package
    '''
    if not package_id:
        try:
            package_id = junoconfig["package_id"]
        except:
            logging.error("must under project dir to run this command when no package_id provided(using project default)")
            exit(1)
    s = deploy_package(base, package_id, kind=0)
    if s:
        logging.error(s)
    else:
        logging.info(f"deploy of package: {package_id} submitted.\nstatus can be viewed by run:\n\tjunocli deploy status")
 
@deploy.command()
@click.argument('package_id', required=False)
@pass_base
@auth
def rollback(base, package_id):
    '''rollback a package to previous version or specific id[optional]
    '''
    token = base.juno_cfg['token']
    if not package_id:
        try:
            package_id = junoconfig["package_id"]
        except:
            logging.error("must under project dir to run this command when no package_id provided(using project default)")
            exit(1)
    res = info_package(base, package_id=package_id)
    if isinstance(res, str):
        logging.error(res)
    else:
        res = list_packages(base, plant=res["plant"], module=res["module"])
        if isinstance(res, str):
            logging.error(str)
        else:
            res.reverse()
            target_idx = -1
            for idx, x in enumerate(res):
                if x["package_id"] == package_id:
                    target_idx = idx +1

            if target_idx < len(res):
                new_id = res[target_idx]["package_id"]
                res = deploy_package(token, new_id , 1)
                if not res:
                    logging.info(f"rollback from {package_id} to {new_id} submitted")
                else:
                    logging.error(res)
            else:
                logging.error(f"no available package to rollback for {package_id}")


def list_deployments(base:CMDBase, plant, module, user, package_id):
    ''' list deoployments
    '''
    api = f"{base.juno_cfg['api']}/deploys"
    params = {}
    params["package_id"] = package_id
    params["plant"] = plant
    params["module"] = module

    if user == "me":
        params["username"] = params["username"] = os.getlogin()
    else:
        params["username"] = user
    logging.info(f"lsit deployments with params:\n{params}")

    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        logging.error(r.text)
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return msg
    else:
        data = r.json()
        data.reverse()
        return data

def list_current_deployed(base:CMDBase, plant:str, module=str):
    api = f"{base.juno_cfg['api']}/deploys/current"
    params = {}
    params["plant"] = plant
    params["module"] = module

    logging.info(f"list current deployments with params:\n{params}")
    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return msg
    else:
        data = r.json()
        data.reverse()
        return data

@deploy.command()
@click.option("-p", "--plant", default="")
@click.option("-m", "--module", default="")
@click.option("-i", "--package_id", default="")
@click.option("-u", "--user", default="")
@pass_base
@auth
def ls(base, plant, module, user, package_id):
    ''' list deoployments
    '''
    res = list_deployments(base, plant, module, user, package_id)
    if isinstance(res, str):
        logging.error(res)
    else:
        logging.info(f"results:\n{json.dumps(res, indent=2)}")


@deploy.command()
@click.argument('plant', required=False)
@click.argument('module', required=False)
@pass_base
@auth
def status(base, plant, module):
    ''' check current deployments status
    '''
    res = list_current_deployed(base, plant=plant, module=module)
    if isinstance(res, str):
        logging.error(res)
    else:
        logging.info(f"results:\n{json.dumps(res, indent=2)}")

def list_pant(base:CMDBase, name:str):
    api = f"{base.juno_cfg['api']}/plants"
    params = {}
    params["name"] = name

    logging.info(f"list plants with params:\n{params}")

    r = requests.get(api, params=params, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild fetch packages "
        if "detail" in r.json():
            msg += r.json()["detail"]
        return msg
    else:
        data = r.json()
        data.reverse()
        return data

@plant.command()
@click.argument('name', required=False, default="")
@pass_base
@auth
def ls(base, name):
    ''' list plant(s)
    '''
    res = list_pant(base, name)
    if isinstance(res, str):
        logging.error(res)
    else:
        logging.info(f"results:\n{json.dumps(res, indent=2)}")

def add_plant(base:CMDBase, name, location, description):
    api = f"{base.juno_cfg['api']}/plant"
    plant = {}
    plant["id"] = 0
    plant["name"] = name
    plant["location"] = location
    plant["description"] = description

    data = {
        "plant": plant
    }

    logging.info(f"add plant with data:\n{data}")

    r = requests.post(api, json=data, headers={"Authorization": f"Bearer {base.juno_cfg['token']}"})
    if r.status_code != 200:
        msg = f"faild add plant "
        if "detail" in r.json():
            msg += f'{r.json()["detail"]}'
        return msg
    else:
        None
    

@plant.command()
@click.argument('name', required=True)
@click.option('-l','--location', required=True)
@click.option('-d', '--description', required=True)
@pass_base
@auth
def add(base, name, location, description):
    ''' add plant
    '''
    res = add_plant(base, name, location, description)
    if isinstance(res, str):
        logging.error(res)